from signalbot import Command
from signalbot import Context as ChatContext

from signalblast.broadcastbot import BroadcasBot
from signalblast.commands_strings import AdminCommandStrings, CommandRegex
from signalblast.utils import triggered


class RemoveAdmin(Command):
    def __init__(self, bot: BroadcasBot) -> None:
        super().__init__()
        self.broadcastbot = bot

    @triggered(CommandRegex.remove_admin)
    async def handle(self, ctx: ChatContext) -> None:
        try:
            subscriber_uuid = ctx.message.source_uuid
            password = self.broadcastbot.message_handler.remove_command_from_message(
                ctx.message.text,
                AdminCommandStrings.remove_admin,
            )

            previous_admin = self.broadcastbot.admin.admin_id
            if await self.broadcastbot.admin.remove(password):
                await self.broadcastbot.reply_with_warn_on_failure(ctx, "Admin has been removed!")
                if previous_admin is not None and subscriber_uuid != previous_admin:
                    msg_to_admin = self.broadcastbot.message_handler.compose_message_to_admin(
                        "You are no longer an admin!",
                        subscriber_uuid,
                    )
                    await self.broadcastbot.send(previous_admin, msg_to_admin)
                self.broadcastbot.logger.info("%s is no longer an admin", previous_admin)
            else:
                await self.broadcastbot.reply_with_warn_on_failure(ctx, "Removing failed: admin password is incorrect!")
                if previous_admin is not None:
                    msg_to_admin = self.broadcastbot.message_handler.compose_message_to_admin(
                        "Tried to remove you as admin",
                        subscriber_uuid,
                    )
                    await self.broadcastbot.send(previous_admin, msg_to_admin)
                self.broadcastbot.logger.warning("Failed password check for remove_admin %s", subscriber_uuid)
        except Exception:
            self.broadcastbot.logger.exception("")



from signalbot import Command
from signalbot import Context as ChatContext

from signalblast.broadcastbot import BroadcasBot
from signalblast.commands_strings import AdminCommandStrings, CommandRegex
from signalblast.utils import triggered


class LastMsgUserUuid(Command):
    def __init__(self, bot: BroadcasBot) -> None:
        super().__init__()
        self.broadcastbot = bot

    @triggered(CommandRegex.last_msg_user_uuid)
    async def handle(self, ctx: ChatContext) -> None:
        try:
            if not await self.broadcastbot.is_user_admin(ctx, AdminCommandStrings.last_msg_user_uuid):
                return

            msg = f"Last message was sent by\n\t{self.broadcastbot.last_msg_user_uuid}"
            await self.broadcastbot.send(self.broadcastbot.admin.admin_id, msg)

            self.broadcastbot.logger.info(msg)
        except Exception:
            self.broadcastbot.logger.exception("")
            try:
                await self.broadcastbot.reply_with_warn_on_failure(ctx, "Failed get UUID")
            except Exception:
                self.broadcastbot.logger.exception("")



from signalbot import Command
from signalbot import Context as ChatContext

from signalblast.broadcastbot import BroadcasBot
from signalblast.commands_strings import CommandRegex, PublicCommandStrings
from signalblast.utils import triggered


class MessageToAdmin(Command):
    def __init__(self, bot: BroadcasBot) -> None:
        super().__init__()
        self.broadcastbot = bot

    @triggered(CommandRegex.msg_to_admin)
    async def handle(self, ctx: ChatContext) -> None:
        try:
            subscriber_uuid = ctx.message.source_uuid
            message = self.broadcastbot.message_handler.remove_command_from_message(
                ctx.message.text,
                PublicCommandStrings.msg_to_admin,
            )

            if self.broadcastbot.admin.admin_id is None:
                await self.broadcastbot.reply_with_warn_on_failure(ctx, "I'm sorry but there are no admins to contact!")
                self.broadcastbot.logger.info("Tried to contact an admin but there is none! %s", subscriber_uuid)
                return

            if subscriber_uuid in self.broadcastbot.banned_users:
                await self.broadcastbot.reply_with_warn_on_failure(ctx, "You are not allowed to contact the admin!")
                self.broadcastbot.logger.info("Banned user %s tried to contact admin", subscriber_uuid)
                return

            msg_to_admin = self.broadcastbot.message_handler.compose_message_to_admin(
                "Sent you message:\n",
                subscriber_uuid,
            )
            msg_to_admin += message
            attachments = self.broadcastbot.message_handler.empty_list_to_none(ctx.message.base64_attachments)
            await self.broadcastbot.send(self.broadcastbot.admin.admin_id, msg_to_admin, base64_attachments=attachments)
            self.broadcastbot.logger.info(
                "Sent message from %s to admin %s",
                subscriber_uuid,
                self.broadcastbot.admin.admin_id,
            )
        except Exception:
            self.broadcastbot.logger.exception("")
            try:
                await self.broadcastbot.reply_with_warn_on_failure(ctx, "Failed to send the message to the admin!")
            except Exception:
                self.broadcastbot.logger.exception("")



from signalbot import Command
from signalbot import Context as ChatContext

from signalblast.broadcastbot import BroadcasBot
from signalblast.commands_strings import AdminCommandStrings, CommandRegex
from signalblast.utils import triggered


class LiftBanSubscriber(Command):
    def __init__(self, bot: BroadcasBot) -> None:
        super().__init__()
        self.broadcastbot = bot

    @triggered(CommandRegex.lift_ban_subscriber)
    async def handle(self, ctx: ChatContext) -> None:
        try:
            user_id = self.broadcastbot.message_handler.remove_command_from_message(
                ctx.message.text,
                AdminCommandStrings.lift_ban_subscriber,
            )

            if not await self.broadcastbot.is_user_admin(ctx, AdminCommandStrings.lift_ban_subscriber):
                return

            if user_id in self.broadcastbot.banned_users:
                await self.broadcastbot.banned_users.remove(user_id)
            else:
                await self.broadcastbot.reply_with_warn_on_failure(
                    ctx,
                    "Could not lift the ban because the user was not banned",
                )
                self.broadcastbot.logger.info("Could not lift the ban of %s because the user was not banned", user_id)
                return

            await self.broadcastbot.send(user_id, "You have banned have been lifted, try subscribing again")
            await self.broadcastbot.reply_with_warn_on_failure(ctx, "Successfully lifted the ban on the user")

            self.broadcastbot.logger.info("Lifted the ban on user %s", user_id)
        except Exception:
            self.broadcastbot.logger.exception("")
            try:
                await self.broadcastbot.reply_with_warn_on_failure(ctx, "Failed lift the ban on the user")
            except Exception:
                self.broadcastbot.logger.exception("")



import functools
from collections.abc import Callable
from logging import Formatter, Logger, getLogger
from logging.handlers import RotatingFileHandler
from pathlib import Path
from re import Pattern
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from signalbot import Context as ChatContext


def _get_rotating_handler(log_file: Path) -> RotatingFileHandler:
    return RotatingFileHandler(str(log_file), mode="a", maxBytes=5 * 1024 * 1024, backupCount=1, encoding=None, delay=0)


def get_logger(name: str | None, log_file: Path, logging_level: int) -> Logger:
    handler = _get_rotating_handler(log_file)
    formatter = Formatter("%(asctime)s %(name)s [%(levelname)s] - %(funcName)s - %(message)s")
    handler.setFormatter(formatter)
    log = getLogger(name)
    log.setLevel(logging_level)
    log.addHandler(handler)
    return log


def redirect_semaphore_logger(log_file: str) -> None:
    formatter = Formatter("%(asctime)s %(name)s [%(levelname)s] - %(message)s")
    handler = _get_rotating_handler(log_file)
    handler.setFormatter(formatter)
    semaphore_log = getLogger("semaphore.bot")
    semaphore_log.addHandler(handler)


def get_code_data_path() -> Path:
    return Path(__file__).parent.absolute() / "data"


def triggered(pattern: Pattern) -> Callable:
    def decorator_triggered(func: Callable) -> Callable:
        @functools.wraps(func)
        async def wrapper_triggered(*args, **kwargs) -> Callable:  # noqa: ANN002, ANN003
            c: ChatContext = args[1]
            text = c.message.text
            if not isinstance(text, str):
                return None

            if pattern.match(text) is None:
                return None

            return await func(*args, **kwargs)

        return wrapper_triggered

    return decorator_triggered



from signalbot import Command
from signalbot import Context as ChatContext

from signalblast.broadcastbot import BroadcasBot
from signalblast.commands_strings import CommandRegex
from signalblast.utils import triggered


class Unsubscribe(Command):
    def __init__(self, bot: BroadcasBot) -> None:
        super().__init__()
        self.broadcastbot = bot

    @triggered(CommandRegex.unsubscribe)
    async def handle(self, ctx: ChatContext) -> None:
        try:
            subscriber_uuid = ctx.message.source_uuid

            if subscriber_uuid not in self.broadcastbot.subscribers:
                await self.broadcastbot.reply_with_warn_on_failure(ctx, "Not subscribed!")
                self.broadcastbot.logger.info("%s tried to unsubscribe but they are not subscribed", subscriber_uuid)
                return

            await self.broadcastbot.subscribers.remove(subscriber_uuid)
            await self.broadcastbot.reply_with_warn_on_failure(ctx, "Successfully unsubscribed!")
            self.broadcastbot.logger.info("%s unsubscribed", subscriber_uuid)
        except Exception:
            self.broadcastbot.logger.exception("")
            try:
                await self.broadcastbot.reply_with_warn_on_failure(ctx, "Could not unsubscribe!")
            except Exception:
                self.broadcastbot.logger.exception("")



from signalbot import Command
from signalbot import Context as ChatContext

from signalblast.broadcastbot import BroadcasBot
from signalblast.commands_strings import AdminCommandStrings, CommandRegex
from signalblast.utils import triggered


class UnsetPing(Command):
    def __init__(self, bot: BroadcasBot) -> None:
        super().__init__()
        self.broadcastbot = bot

    @triggered(CommandRegex.unset_ping)
    async def handle(self, ctx: ChatContext) -> None:
        try:
            if not await self.broadcastbot.is_user_admin(ctx, AdminCommandStrings.unset_ping):
                return

            if self.broadcastbot.ping_job is None:
                await ctx.reply("Cannot unset because ping was not set!")
                return

            self.broadcastbot.scheduler.remove_job(self.broadcastbot.ping_job.id)
            self.broadcastbot.ping_job = None
            await ctx.reply("Ping unset!")
        except Exception:
            self.broadcastbot.logger.exception("")
            try:
                await self.broadcastbot.reply_with_warn_on_failure(ctx, "Failed to unset ping")
            except Exception:
                self.broadcastbot.logger.exception("")



from pathlib import Path

from signalblast.commands_strings import AdminCommandArgs, AdminCommandStrings, PublicCommandStrings


class MessageHandler:
    def __init__(self, attachments_folder: Path) -> None:
        self.attachments_folder = attachments_folder

    @staticmethod
    def remove_command_from_message(message: str | None, command: str) -> str | None:
        if message == "" or message is None:
            return None
        message = message.replace(command, "", 1).strip()
        if message == "":
            return None
        return message

    def empty_list_to_none(self, attachments: list[str] | None) -> list | None:
        if attachments == []:
            return None

        return attachments

    def delete_attachments(self, attachments: list[str] | None, link_previews: list[str] | None) -> None:
        if attachments is not None:
            for attachment in attachments:
                (self.attachments_folder / attachment).unlink()

        if link_previews is not None:
            error_msg = "Link previews are not implemented yet"
            for _ in link_previews:
                raise NotImplementedError(error_msg)

    @staticmethod
    def _compose_help_message(*, add_admin_commands: bool) -> str:
        def _add_commands(message: str) -> str:
            for command_str in PublicCommandStrings:
                message += "\t" + command_str + "\n"
            return message

        def _add_admin_commands(message: str) -> str:
            for command_str, command_arg in zip(AdminCommandStrings, AdminCommandArgs, strict=False):
                message += "\t" + command_str + " " + command_arg + "\n"
            return message

        message = _add_commands("")
        if add_admin_commands:
            message = _add_admin_commands(message)
        return message

    @staticmethod
    def compose_help_message(*, add_admin_commands: bool = False, is_help: bool = True) -> str:
        message = MessageHandler._compose_help_message(add_admin_commands=add_admin_commands)
        if is_help:
            return "I'm happy to help! This are the commands that you can use:\n\n" + message
        message = "I'm sorry, I didn't understand you but I understand the following commands:\n\n" + message
        message += "\nPlease try again"
        return message

    @staticmethod
    def compose_must_subscribe_message() -> str:
        message = "To be able to send messages you must sign up.\n"
        message += "Please sign up by sending:\n"
        message += f"\t{PublicCommandStrings.subscribe}\n"
        message += "and try again after that."
        return message

    @staticmethod
    def compose_message_to_admin(message: str, user: str | None) -> str:
        header = "***Admin***\n"
        if user is not None:
            header += user + "\n"
        return header + message

    @staticmethod
    def compose_welcome_message(default_message: str | None) -> str:
        if default_message is None:
            return "Subscription successful!"
        return default_message



import bcrypt

from signalblast.utils import get_code_data_path


class Admin:
    save_path = get_code_data_path() / "admin.txt"

    def __init__(self) -> None:
        self.admin_id: str = None
        self._hashed_password: str = None

    @classmethod
    async def create(cls, admin_password: str | None) -> None:
        self = Admin()
        self.admin_id = None
        await self.set_hashed_password(admin_password)
        return self

    def get_hashed_password(self) -> str:
        return self._hashed_password

    async def set_hashed_password(self, password: str) -> None:
        if password is None:
            self._hashed_password = b""
        else:
            self._hashed_password = bcrypt.hashpw(password.encode(), bcrypt.gensalt())
        await self.save_to_file()

    async def add(self, admin_id: str, admin_password: str | None) -> bool:
        if admin_password is None:
            return False

        if bcrypt.checkpw(admin_password.encode(), self.get_hashed_password()):
            self.admin_id = admin_id
            await self.save_to_file()
            return True
        return False

    async def remove(self, admin_password: str | None) -> bool:
        if admin_password is None:
            return False

        if bcrypt.checkpw(admin_password.encode(), self.get_hashed_password()):
            self.admin_id = None
            await self.save_to_file()
            return True
        return False

    async def save_to_file(self) -> None:
        with self.save_path.open("w") as f:
            if self.admin_id is None:
                f.write("\n")
            else:
                f.write(self.admin_id + "\n")
            f.write(self.get_hashed_password().decode())

    @staticmethod
    async def _load_from_file() -> "Admin":
        admin = Admin()
        with Admin.save_path.open() as f:
            admin.admin_id = f.readline().rstrip()
            admin._hashed_password = f.readline().encode()  # noqa: SLF001

        if admin.admin_id == "":
            admin.admin_id = None

        return admin

    @staticmethod
    async def load_from_file(admin_password: str | None) -> "Admin":
        if not Admin.save_path.exists():
            return await Admin.create(admin_password)

        admin = await Admin._load_from_file()
        # Overwrite the password in the file, if no password was given assume we want to keep the one from the file
        if admin_password is not None:
            await admin.set_hashed_password(admin_password)
        return admin



import asyncio
import contextlib
import random
from collections import defaultdict
from re import Pattern

from signalbot import Command
from signalbot import Context as ChatContext

from signalblast.broadcastbot import BroadcasBot
from signalblast.commands_strings import CommandRegex, PublicCommandStrings


class Broadcast(Command):
    MAX_FAILED_MSGS = 10

    def __init__(self, bot: BroadcasBot) -> None:
        super().__init__()
        self.broadcastbot = bot
        self.subscribers_num_fails: dict[str, int] = defaultdict(lambda: 0)

    def is_valid_command(self, message: str, invalid_command: Pattern) -> bool:
        return any(regex != invalid_command and regex.search(message) is not None for regex in CommandRegex)

    async def check_send_tasks_results(self, send_tasks: list[asyncio.Task | None]) -> int:
        num_broadcasts = 0
        for send_task, subscriber in zip(send_tasks, self.broadcastbot.subscribers, strict=False):
            if send_task is not None:
                if send_task.exception() is None:
                    num_broadcasts += 1
                    self.subscribers_num_fails.pop(subscriber, None)
                    self.broadcastbot.logger.info("Message successfully sent to %s", subscriber)
                else:
                    self.subscribers_num_fails[subscriber] += 1
                    self.broadcastbot.logger.warning("Could not send message to %s", subscriber)

        subscribers_to_remove = []
        for subscriber, num_fails in self.subscribers_num_fails.items():
            if num_fails >= Broadcast.MAX_FAILED_MSGS:
                subscribers_to_remove.append(subscriber)
                await self.broadcastbot.subscribers.remove(subscriber)

                remove_message = "The bot is having problems sending you messages. "
                remove_message += "You have been removed from the list. "
                remove_message += "Please update signal, remove old linked devices and try subscribing again."
                with contextlib.suppress(Exception):
                    # Most likely will fail to send the message but try anyway
                    await self.broadcastbot.send(subscriber, remove_message)

        for subscriber in subscribers_to_remove:
            del self.subscribers_num_fails[subscriber]

        return num_broadcasts

    async def broadcast(self, ctx: ChatContext) -> None:
        num_broadcasts = 0
        num_subscribers = -1
        attachments_deleted = False
        send_tasks_checked = False
        send_tasks: list[asyncio.Task | None] = []

        try:
            subscriber_uuid = ctx.message.source_uuid
            if subscriber_uuid in self.broadcastbot.banned_users:
                await self.broadcastbot.send(subscriber_uuid, "This number is not allowed to send messages")
                self.broadcastbot.logger.info("%s tried to broadcast but they are banned", subscriber_uuid)
                return

            if subscriber_uuid not in self.broadcastbot.subscribers:
                await self.broadcastbot.send(subscriber_uuid, self.broadcastbot.must_subscribe_message)
                self.broadcastbot.logger.info("%s tried to broadcast but they are not subscribed", subscriber_uuid)
                return

            num_subscribers = len(self.broadcastbot.subscribers)

            # await ctx.message.mark_read()
            message = self.broadcastbot.message_handler.remove_command_from_message(
                ctx.message.text,
                PublicCommandStrings.broadcast,
            )
            attachments = self.broadcastbot.message_handler.empty_list_to_none(ctx.message.base64_attachments)
            # link_previews = self.broadcastbot.message_handler.empty_list_to_none(ctx.message.data_message.previews)

            if message is None and attachments is None:
                return

            if message is None:
                message = ""

            # Broadcast message to all subscribers.
            send_tasks: list[asyncio.Task | None] = [None] * num_subscribers

            for i, subscriber in enumerate(self.broadcastbot.subscribers):
                send_tasks[i] = asyncio.create_task(
                    self.broadcastbot.send(subscriber, message, base64_attachments=attachments),
                )

                # Avoid rate limiting by waiting a random time between messages
                await asyncio.sleep(random.uniform(0.5, 1))  # noqa: S311

            await asyncio.wait(send_tasks)

            num_broadcasts = await self.check_send_tasks_results(send_tasks)
            send_tasks_checked = True

            attachments_filenames = self.broadcastbot.message_handler.empty_list_to_none(
                ctx.message.attachments_filenames,
            )
            self.broadcastbot.message_handler.delete_attachments(attachments_filenames, link_previews=None)
            attachments_deleted = True

            await self.broadcastbot.reply_with_warn_on_failure(ctx, f"Message sent to {num_broadcasts - 1} people")

            self.broadcastbot.last_msg_user_uuid = subscriber_uuid
        except Exception:
            self.broadcastbot.logger.exception("")
            try:
                if send_tasks_checked is False:
                    num_broadcasts = await self.check_send_tasks_results(send_tasks)

                error_str = "Something went wrong when sending the message"
                error_str += f", it was only sent to {num_broadcasts - 1} out of {num_subscribers - 1} people"
                error_str += ", please contact the admin if the problem persists"
                await self.broadcastbot.reply_with_warn_on_failure(ctx, error_str)

                if attachments_deleted is False:
                    attachments_filenames = self.broadcastbot.message_handler.empty_list_to_none(
                        ctx.message.attachments_filenames,
                    )
                    self.broadcastbot.message_handler.delete_attachments(attachments_filenames, link_previews=None)
            except Exception:
                self.broadcastbot.logger.exception("")

    async def handle(self, ctx: ChatContext) -> None:
        message = ctx.message.text
        subscriber_uuid = ctx.message.source_uuid
        if message is None:
            if ctx.message.base64_attachments == []:
                self.broadcastbot.logger.info("Received reaction, sticker or similar from %s", subscriber_uuid)
                return

            # Only attachment, assume the user wants to forward that
            self.broadcastbot.logger.info("Received a file from %s, broadcasting!", subscriber_uuid)
            await self.broadcast(ctx)
            return

        if self.is_valid_command(message, invalid_command=CommandRegex.broadcast):
            return

        # By default broadcast all the messages
        await self.broadcast(ctx)



from signalbot import Command
from signalbot import Context as ChatContext

from signalblast.broadcastbot import BroadcasBot
from signalblast.commands_strings import AdminCommandStrings, CommandRegex
from signalblast.utils import triggered


class SetPing(Command):
    def __init__(self, bot: BroadcasBot) -> None:
        super().__init__()
        self.broadcastbot = bot

    async def _send_ping(self, ctx: ChatContext) -> None:
        try:
            await self.broadcastbot.reply_with_warn_on_failure(ctx, "Ping")
        except Exception:
            self.broadcastbot.logger.exception("")
            try:
                await self.broadcastbot.reply_with_warn_on_failure(ctx, "Failed to send ping")
            except Exception:
                self.broadcastbot.logger.exception("")

    @triggered(CommandRegex.set_ping)
    async def handle(self, ctx: ChatContext) -> None:
        try:
            ping_time = self.broadcastbot.message_handler.remove_command_from_message(
                ctx.message.text,
                AdminCommandStrings.set_ping,
            )

            if not await self.broadcastbot.is_user_admin(ctx, AdminCommandStrings.set_ping):
                return

            if self.broadcastbot.ping_job is not None:
                self.broadcastbot.scheduler.remove_job(self.broadcastbot.ping_job.id)
                self.broadcastbot.logger.info("Unset old ping job")
                await self.broadcastbot.reply_with_warn_on_failure(ctx, "Unset old ping job")

            self.broadcastbot.ping_job = self.broadcastbot.scheduler.add_job(
                self._send_ping,
                "interval",
                seconds=int(ping_time),
                args=[ctx],
            )

            await self.broadcastbot.reply_with_warn_on_failure(ctx, f"Ping set every {ping_time} seconds")
            self.broadcastbot.logger.info("Ping set every %s seconds", ping_time)
        except Exception:
            self.broadcastbot.logger.exception("")
            try:
                await self.broadcastbot.reply_with_warn_on_failure(ctx, "Failed set ping")
            except Exception:
                self.broadcastbot.logger.exception("")



from signalbot import Command
from signalbot import Context as ChatContext

from signalblast.broadcastbot import BroadcasBot
from signalblast.commands_strings import AdminCommandStrings, CommandRegex
from signalblast.utils import triggered


class BanSubscriber(Command):
    def __init__(self, bot: BroadcasBot) -> None:
        super().__init__()
        self.broadcastbot = bot

    @triggered(CommandRegex.ban_subscriber)
    async def handle(self, ctx: ChatContext) -> None:
        try:
            user_id = self.broadcastbot.message_handler.remove_command_from_message(
                ctx.message.text,
                AdminCommandStrings.ban_subscriber,
            )

            if not await self.broadcastbot.is_user_admin(ctx, AdminCommandStrings.ban_subscriber):
                return

            if user_id in self.broadcastbot.subscribers:
                await self.broadcastbot.subscribers.remove(user_id)

            user_phone_number = self.broadcastbot.subscribers.get_phone_number(user_id)
            await self.broadcastbot.banned_users.add(user_id, user_phone_number)

            await self.broadcastbot.send(user_id, "You have been banned")
            await self.broadcastbot.reply_with_warn_on_failure(ctx, "Successfully banned user")

            self.broadcastbot.logger.info("Banned user %s", user_id)
        except Exception:
            self.broadcastbot.logger.exception("")
            try:
                await self.broadcastbot.reply_with_warn_on_failure(ctx, "Failed to ban user")
            except Exception:
                self.broadcastbot.logger.exception("")



#!/usr/bin/env python

import argparse
import asyncio
import logging
import os
from pathlib import Path

from signalblast.broadcastbot import BroadcasBot
from signalblast.commands import (
    AddAdmin,
    BanSubscriber,
    Broadcast,
    DisplayHelp,
    LastMsgUserUuid,
    LiftBanSubscriber,
    MessageFromAdmin,
    MessageToAdmin,
    RemoveAdmin,
    SetPing,
    Subscribe,
    UnsetPing,
    Unsubscribe,
)
from signalblast.utils import get_code_data_path, get_logger

LOGGING_LEVEL = logging.WARNING
LOG_TO_FILE = False
if LOG_TO_FILE:
    get_logger(None, Path("signalbot.log"), LOGGING_LEVEL)
else:
    logging.getLogger().setLevel(LOGGING_LEVEL)

logging.getLogger("apscheduler").setLevel(LOGGING_LEVEL)


async def initialise_bot(  # noqa: PLR0913 Too many arguments in function definition
    signal_service: str,
    phone_number: str,
    admin_pass: str,
    expiration_time: int,
    signal_data_path: Path,
    welcome_message: str | None = None,
    storage: dict[str, str] | None = None,
) -> BroadcasBot:
    config = {
        "signal_service": signal_service,
        "phone_number": phone_number,
        "storage": storage,
    }

    get_code_data_path().mkdir(parents=True, exist_ok=True)
    if LOG_TO_FILE:
        logger = get_logger("signalblast", Path("signalblast.log"), LOGGING_LEVEL)
    else:
        logger = logging.getLogger("signalblast")
        logger.setLevel(LOGGING_LEVEL)

    bot = BroadcasBot(config)
    await bot.load_data(
        logger=logger,
        admin_pass=admin_pass,
        expiration_time=expiration_time,
        signal_data_path=signal_data_path,
        welcome_message=welcome_message,
    )

    bot.register(Subscribe(bot=bot))
    bot.register(Unsubscribe(bot=bot))
    bot.register(Broadcast(bot=bot))
    bot.register(DisplayHelp(bot=bot))
    bot.register(AddAdmin(bot=bot))
    bot.register(RemoveAdmin(bot=bot))
    bot.register(BanSubscriber(bot=bot))
    bot.register(LiftBanSubscriber(bot=bot))
    bot.register(SetPing(bot=bot), contacts=False, groups=True)
    bot.register(UnsetPing(bot=bot), contacts=False, groups=True)
    bot.register(MessageToAdmin(bot=bot))
    bot.register(MessageFromAdmin(bot=bot))
    bot.register(LastMsgUserUuid(bot=bot))
    return bot


if __name__ == "__main__":
    one_week = 60 * 60 * 24 * 7  # Number of seconds in a week

    args_parser = argparse.ArgumentParser()
    args_parser.add_argument(
        "--admin_pass",
        type=str,
        help="the password to add or remove admins",
        default=os.environ.get("SIGNALBLAST_PASSWORD"),
    )
    args_parser.add_argument(
        "--expiration_time",
        type=int,
        default=os.environ.get("SIGNALBLAST_EXPIRATION_TIME", one_week),
        help="the expiration time for the chats in seconds",
    )
    args_parser.add_argument(
        "--signal_data_path",
        type=Path,
        default=os.environ.get("SIGNAL_DATA_PATH", Path.home() / (".local/share/signal-api/")),
        help="the path to the folder containig the signal api data",
    )

    args_parser.add_argument(
        "--signal_service",
        type=str,
        default=os.environ.get("SIGNAL_SERVICE", "localhost:8080"),
        help="the address of the signal cli rest api",
    )

    args_parser.add_argument(
        "--phone_number",
        type=str,
        default=os.environ.get("SIGNALBLAST_PHONE_NUMBER"),
        help="the phone number of the bot",
    )

    args_parser.add_argument(
        "--welcome_message",
        type=str,
        default=os.environ.get("SIGNALBLAST_WELCOME_MESSAGE"),
        help="the initial message that the user receives",
    )

    args = args_parser.parse_args()

    if args.phone_number is None:
        value_error_msg = "The bot phone number is not set"
        raise ValueError(value_error_msg)

    loop = asyncio.get_event_loop()
    bot = loop.run_until_complete(
        initialise_bot(
            signal_service=args.signal_service,
            phone_number=args.phone_number,
            signal_data_path=args.signal_data_path,
            admin_pass=args.admin_pass,
            expiration_time=args.expiration_time,
            welcome_message=args.welcome_message,
        ),
    )
    bot.start()



from collections.abc import Callable
from logging import Logger
from pathlib import Path
from typing import TYPE_CHECKING

from apscheduler.schedulers.asyncio import AsyncIOScheduler
from signalbot import Command, Message, SignalBot
from signalbot import Context as ChatContext

from signalblast.admin import Admin
from signalblast.message_handler import MessageHandler
from signalblast.users import Users
from signalblast.utils import get_code_data_path

if TYPE_CHECKING:
    from apscheduler.job import Job


class BroadcasBot:
    subscribers_data_path = get_code_data_path() / "subscribers.csv"
    banned_users_data_path = get_code_data_path() / "banned_users.csv"

    def __init__(self, config: dict) -> None:
        self._bot = SignalBot(config)
        self.ping_job: Job | None = None
        self.last_msg_user_uuid: str | None = None

        # Type hint the other attributes that will get defined in load_data
        self.subscribers: Users
        self.banned_users: Users
        self.admin: Admin
        self.message_handler: MessageHandler
        self.help_message: str
        self.wrong_command_message: str
        self.admin_help_message: str
        self.admin_wrong_command_message: str
        self.must_subscribe_message: str
        self.logger: Logger
        self.expiration_time: int
        self.welcome_message: str

    async def send(  # noqa: PLR0913 Too many arguments in function definition
        self,
        receiver: str,
        text: str,
        base64_attachments: list | None = None,
        quote_author: str | None = None,
        quote_mentions: list | None = None,
        quote_message: str | None = None,
        quote_timestamp: str | None = None,
        mentions: list | None = None,
        text_mode: str | None = None,
        listen: bool = False,  # noqa: FBT001, FBT002
    ) -> str:
        return await self._bot.send(
            receiver=receiver,
            text=text,
            base64_attachments=base64_attachments,
            quote_author=quote_author,
            quote_mentions=quote_mentions,
            quote_message=quote_message,
            quote_timestamp=quote_timestamp,
            mentions=mentions,
            text_mode=text_mode,
            listen=listen,
        )

    def register(
        self,
        command: Command,
        contacts: list[str] | bool | None = True,  # noqa: FBT002
        groups: list[str] | bool | None = False,  # noqa: FBT002
        f: Callable[[Message], bool] | None = None,
    ) -> None:
        self._bot.register(command=command, contacts=contacts, groups=groups, f=f)

    def start(self) -> None:
        self._bot.start()

    @property
    def scheduler(self) -> AsyncIOScheduler:
        return self._bot.scheduler

    async def load_data(
        self,
        logger: Logger,
        admin_pass: str | None,
        expiration_time: int | None,
        signal_data_path: Path,
        welcome_message: str | None = None,
    ) -> None:
        self.subscribers = await Users.load_from_file(self.subscribers_data_path)
        self.banned_users = await Users.load_from_file(self.banned_users_data_path)

        self.admin = await Admin.load_from_file(admin_pass)
        self.message_handler = MessageHandler(signal_data_path / "attachments")

        self.help_message = self.message_handler.compose_help_message(is_help=True)
        self.wrong_command_message = self.message_handler.compose_help_message(is_help=False)
        self.admin_help_message = self.message_handler.compose_help_message(add_admin_commands=True)
        self.admin_wrong_command_message = self.message_handler.compose_help_message(
            add_admin_commands=True,
            is_help=False,
        )
        self.welcome_message = self.message_handler.compose_welcome_message(welcome_message)

        self.must_subscribe_message = self.message_handler.compose_must_subscribe_message()

        self.expiration_time = expiration_time

        self.logger = logger
        self.logger.debug("BotAnswers is initialised")

    async def reply_with_warn_on_failure(self, ctx: ChatContext, message: str) -> bool:
        if await ctx.reply(message):
            return True
        self.logger.warning("Could not send message to %s", ctx.message.source_uuid)
        return False

    async def is_user_admin(self, ctx: ChatContext, command: str) -> bool:
        subscriber_uuid = ctx.message.source_uuid
        if self.admin.admin_id is None:
            await self.reply_with_warn_on_failure(ctx, "I'm sorry but there are no admins")
            self.logger.info("Tried to %s but there are no admins! %s", command, subscriber_uuid)
            return False

        if self.admin.admin_id != subscriber_uuid:
            await self.reply_with_warn_on_failure(ctx, "I'm sorry but you are not an admin")
            msg_to_admin = self.message_handler.compose_message_to_admin(f"Tried to {command}", subscriber_uuid)
            await self.send(self.admin.admin_id, msg_to_admin)
            self.logger.info("%s tried to %s but admin is %s", subscriber_uuid, command, self.admin.admin_id)
            return False

        return True



from signalbot import Command
from signalbot import Context as ChatContext

from signalblast.broadcastbot import BroadcasBot
from signalblast.commands_strings import CommandRegex, PublicCommandStrings
from signalblast.utils import triggered


class DisplayHelp(Command):
    def __init__(self, bot: BroadcasBot) -> None:
        super().__init__()
        self.broadcastbot = bot

    def _get_help_message(self, input_message: str, subscriber_uuid: str) -> str:
        is_wrong_command = not input_message.startswith(PublicCommandStrings.help)

        if subscriber_uuid != self.broadcastbot.admin.admin_id:
            if is_wrong_command:
                return self.broadcastbot.wrong_command_message
            return self.broadcastbot.help_message
        if is_wrong_command:
            return self.broadcastbot.admin_wrong_command_message
        return self.broadcastbot.admin_help_message

    @triggered(CommandRegex.help)
    async def handle(self, ctx: ChatContext) -> None:
        try:
            subscriber_uuid = ctx.message.source_uuid
            message = ctx.message.text

            help_message = self._get_help_message(message, subscriber_uuid)

            await self.broadcastbot.reply_with_warn_on_failure(ctx, help_message)
            self.broadcastbot.logger.info("Sent help message to %s", subscriber_uuid)
        except Exception:
            # if isinstance(e, StopPropagation):
            #     raise
            self.broadcastbot.logger.exception("")



from signalblast.commands.add_admin import AddAdmin  # noqa: F401
from signalblast.commands.ban_subscriber import BanSubscriber  # noqa: F401
from signalblast.commands.broadcast import Broadcast  # noqa: F401
from signalblast.commands.display_help import DisplayHelp  # noqa: F401
from signalblast.commands.last_msg_user_uuid import LastMsgUserUuid  # noqa: F401
from signalblast.commands.lift_ban_subscriber import LiftBanSubscriber  # noqa: F401
from signalblast.commands.message_from_admin import MessageFromAdmin  # noqa: F401
from signalblast.commands.message_to_admin import MessageToAdmin  # noqa: F401
from signalblast.commands.remove_admin import RemoveAdmin  # noqa: F401
from signalblast.commands.set_ping import SetPing  # noqa: F401
from signalblast.commands.subscribe import Subscribe  # noqa: F401
from signalblast.commands.unset_ping import UnsetPing  # noqa: F401
from signalblast.commands.unsubscribe import Unsubscribe  # noqa: F401



from __future__ import annotations

import csv
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Iterator
    from pathlib import Path


class Users:
    _uuid_str = "uuid"
    _phone_number_str = "phone_number"

    def __init__(self, save_path: Path) -> None:
        self.save_path = save_path
        self.data: dict[str, str | None] = {}

    async def add(self, uuid: str, phone_number: str | None) -> None:
        self.data[uuid] = phone_number
        await self.save_to_file()

    async def remove(self, uuid: str) -> None:
        del self.data[uuid]
        await self.save_to_file()

    async def save_to_file(self) -> None:
        with self.save_path.open("w") as f:
            csv_writer = csv.DictWriter(f, fieldnames=[self._uuid_str, self._phone_number_str])
            csv_writer.writeheader()
            for uuid, phone_number in self.data.items():
                csv_writer.writerow({self._uuid_str: uuid, self._phone_number_str: phone_number})

    @staticmethod
    async def _load_from_file(save_path: Path) -> Users:
        users = Users(save_path)
        with save_path.open() as f:
            csv_reader = csv.DictReader(f)
            for line in csv_reader:
                users.data[line[Users._uuid_str]] = line[Users._phone_number_str]
        return users

    def get_phone_number(self, uuid: str) -> str | None:
        return self.data.get(uuid)

    def __iter__(self) -> Iterator[str | None]:
        yield from self.data

    def __contains__(self, uuid: str) -> bool:
        return uuid in self.data

    def __len__(self) -> int:
        return len(self.data)

    @staticmethod
    async def load_from_file(save_path: Path) -> Users:
        if not save_path.exists():
            return Users(save_path)

        return await Users._load_from_file(save_path)



from signalbot import Command
from signalbot import Context as ChatContext

from signalblast.broadcastbot import BroadcasBot
from signalblast.commands_strings import AdminCommandStrings, CommandRegex
from signalblast.utils import triggered


class MessageFromAdmin(Command):
    def __init__(self, bot: BroadcasBot) -> None:
        super().__init__()
        self.broadcastbot = bot

    @triggered(CommandRegex.msg_from_admin)
    async def handle(self, ctx: ChatContext) -> None:
        try:
            message = self.broadcastbot.message_handler.remove_command_from_message(
                ctx.message.text,
                AdminCommandStrings.msg_from_admin,
            )

            if not await self.broadcastbot.is_user_admin(ctx, AdminCommandStrings.msg_from_admin):
                return

            user_id, message = message.split(" ", 1)

            if user_id not in self.broadcastbot.subscribers:
                if " " in message:
                    confirmation, message = message.split(" ", 1)
                else:
                    confirmation = None
                if confirmation != "!force":
                    warn_message = "User is not in subscribers list, use !reply <uuid> !force to message them"
                    await self.broadcastbot.reply_with_warn_on_failure(ctx, warn_message)
                    return

            message = "Admin: " + message
            attachments = self.broadcastbot.message_handler.empty_list_to_none(ctx.message.base64_attachments)

            await self.broadcastbot.send(user_id, message, base64_attachments=attachments)
            self.broadcastbot.logger.info(
                "Sent message from admin %s to user %s",
                self.broadcastbot.admin.admin_id,
                user_id,
            )
        except Exception:
            self.broadcastbot.logger.exception("")
            try:
                await self.broadcastbot.reply_with_warn_on_failure(ctx, "Failed to send the message to the user!")
            except Exception:
                self.broadcastbot.logger.exception("")



from signalbot import Command
from signalbot import Context as ChatContext

from signalblast.broadcastbot import BroadcasBot
from signalblast.commands_strings import CommandRegex
from signalblast.utils import triggered


class Subscribe(Command):
    def __init__(self, bot: BroadcasBot) -> None:
        super().__init__()
        self.broadcastbot = bot

    async def subscribe(self, ctx: ChatContext, *, verbose: bool = False) -> None:
        try:
            subscriber_uuid = ctx.message.source_uuid
            if subscriber_uuid in self.broadcastbot.subscribers:
                if verbose:
                    await self.broadcastbot.reply_with_warn_on_failure(ctx, "Already subscribed!")
                    self.broadcastbot.logger.info("Already subscribed!")
                return

            if subscriber_uuid in self.broadcastbot.banned_users:
                if verbose:
                    await self.broadcastbot.reply_with_warn_on_failure(ctx, "This number is not allowed to subscribe")
                    self.broadcastbot.logger.info("%s was not allowed to subscribe", subscriber_uuid)
                return

            await self.broadcastbot.subscribers.add(subscriber_uuid, ctx.message.source_number)
            await self.broadcastbot.reply_with_warn_on_failure(ctx, self.broadcastbot.welcome_message)

            # TODO(Era): Enable when the backend supports this # noqa: TD003
            # if self.broadcastbot.expiration_time is not None:
            #     await ctx.bot.set_expiration(subscriber_uuid, self.broadcastbot.expiration_time)

            self.broadcastbot.logger.info("%s subscribed", subscriber_uuid)
        except Exception:
            self.broadcastbot.logger.exception("")
            try:
                await self.broadcastbot.reply_with_warn_on_failure(ctx, "Could not subscribe!")
            except Exception:
                self.broadcastbot.logger.exception("")

    @triggered(CommandRegex.subscribe)
    async def handle(self, ctx: ChatContext) -> None:
        await Subscribe.subscribe(self, ctx, verbose=True)



import inspect
import re
from collections.abc import Iterator
from dataclasses import dataclass


class _IterableDataClass:
    _public_attr = None

    @classmethod
    def _get_public_attributes(cls) -> None:
        cls._public_attr: list[str] = []
        for attr, value in inspect.getmembers(cls):
            if not (attr.startswith("_") or inspect.ismethod(value)):
                cls._public_attr.append(value)

    def __iter__(self) -> Iterator[str]:
        if self._public_attr is None:
            self._get_public_attributes()

        yield from self._public_attr


@dataclass
class _PublicCommandStrings(_IterableDataClass):
    subscribe = "!subscribe"
    unsubscribe = "!unsubscribe"
    broadcast = "!broadcast"
    msg_to_admin = "!admin"
    help = "!help"


PublicCommandStrings = _PublicCommandStrings()


@dataclass
class _AdminCommandStrings(_IterableDataClass):
    add_admin = "!add admin"
    remove_admin = "!remove admin"
    msg_from_admin = "!reply"
    ban_subscriber = "!ban"
    lift_ban_subscriber = "!lift ban"
    set_ping = "!set ping"
    unset_ping = "!unset ping"
    last_msg_user_uuid = "!last msg user uuid"


AdminCommandStrings = _AdminCommandStrings()


@dataclass
class _AdminCommandArgs(_IterableDataClass):
    add_admin = "<password>"
    remove_admin = "<password>"
    msg_from_admin = "<user id>"
    ban_subscriber = "<user id>"
    lift_ban_subscriber = "<user id>"
    set_ping = "<time>"
    unset_ping = ""
    last_msg_user_uuid = ""


AdminCommandArgs = _AdminCommandArgs()


def _begings_with(in_str: str) -> str:
    return "^(" + in_str + ")"


@dataclass
class _CommandRegex(_IterableDataClass):
    subscribe = re.compile(_begings_with(PublicCommandStrings.subscribe))
    unsubscribe = re.compile(_begings_with(PublicCommandStrings.unsubscribe))
    broadcast = re.compile(_begings_with(PublicCommandStrings.broadcast))
    add_admin = re.compile(_begings_with(AdminCommandStrings.add_admin))
    remove_admin = re.compile(_begings_with(AdminCommandStrings.remove_admin))
    msg_to_admin = re.compile(_begings_with(PublicCommandStrings.msg_to_admin))
    msg_from_admin = re.compile(_begings_with(AdminCommandStrings.msg_from_admin))
    ban_subscriber = re.compile(_begings_with(AdminCommandStrings.ban_subscriber))
    lift_ban_subscriber = re.compile(_begings_with(AdminCommandStrings.lift_ban_subscriber))
    set_ping = re.compile(_begings_with(AdminCommandStrings.set_ping))
    unset_ping = re.compile(_begings_with(AdminCommandStrings.unset_ping))
    help = re.compile(_begings_with(PublicCommandStrings.help))
    last_msg_user_uuid = re.compile(_begings_with(AdminCommandStrings.last_msg_user_uuid))


CommandRegex = _CommandRegex()
